#!/usr/bin/env python

def byteswap(rom):
    from array import array
    a = array("H", rom)
    a.byteswap()
    return a.tobytes()

def byteorder(rom):
    # Lazy test for byteswapping; let me know about any fringe cases.
    if rom[0]==128:
        return 'big'
    elif rom[1]==128:
        return 'little'
    else:
        raise(ValueError, "Unable to determine byteorder from initial PI settings.")

def header(data):
    """Extracts header data from an unbyteswapped (presumably) N64 ROM.
    IMPORTANT NOTE: currently skipping initialization fields. (ie. PI settings, boot address, clock speed, etc.)

    Returns a namedtuple with the following fields:
        title   cp932 string with encoded filename
        frmt    'N' for carts, 'D' for disk, 'C'/'E' for cart/expansion titles
        name    2-char name; (mostly) unique for each title
        region  char designating intended region
        version integer value, starting at 0, for each release version
    """
    from collections import namedtuple
##        import io
##        if isinstance(data, io.IOBase):
##            data = data.read(64)

    title = data[32:52].decode(encoding='cp932',errors='replace').strip()
    # get ID
    frmt=data[59:60].decode(errors="replace")
    ID  =data[60:62].decode(errors="replace")
    reg =data[62:63].decode(errors="replace")
    ver =data[63]
    clck=int.from_bytes(data[4:8],  byteorder='big') & ~0xF
    pc  =int.from_bytes(data[8:12], byteorder='big')
    rls =int.from_bytes(data[12:16],byteorder='big')
    nt  =namedtuple('ID', ['initPI', 'clock', 'PC', 'release', 'title', 'frmt', 'name', 'region', 'version'])
    pi  =namedtuple('PI', ['release', 'pagesize', 'pulsewidth', 'latency'])
    p = pi(data[1]>>4, data[1]&0xF, data[2], data[3])
    return nt(p, clck, pc, rls, title, frmt, ID, reg, ver)

def cic_round(mem):
    """[mem] should be an array of 16 unsigned long values.
    Implementation of an algorithm presented in ccsfp221-kammerstetter.pdf

    First two nibbles are from console, rest are hardware-randomized by free pin.
    Called 3 times on array.  7th nibble dictates how many bits are sent to console.
    Sends least significant bit of each nibble until #bits met."""
    x = mem[15]
    while True:
        mem[1] += x + 1

        b = 7 if (15 - mem[3]) > mem[2] else 6
        mem[b - 3] += mem[3]
        mem[3] += mem[2] + 1
        mem[2] = -(mem[2] + mem[1] + 1)

        a = mem[b - 1]
        if mem[b - 2] > 7:
            mem[b - 1] = 0
        mem[b - 1] += mem[b - 2] + 8
        mem[b - 2] += mem[b - 3]

        while True:
            a += mem[b] + 1
            mem[b] = a
            b += 1
            if b == 0:
                break
        x -= 1
        if x == 15:
            break
    return mem

def cic_6105(challenge, rsp):
    """Simulation of the 6105 challenge algorithm.
    Only practical use is generating values for file encode/decode."""
    lut0 = (4, 7, 10, 7, 14, 5, 14, 1, 12, 15, 8, 15, 6, 3, 6, 9)
    lut1 = (4, 1, 10, 7, 14, 5, 14, 1, 12, 9, 8, 5, 6, 3, 12, 9)
    # Copy to working lut.
    lut = lut0
    key = 11
    for i in range(len(challenge)-2):
        rsp[i] = (key + 5 * challenge[i]) & 0xF
        key = lut[rsp[i]]
        sgn = (rsp[i]>>3) & 1
        mag = (~rsp[i] if sgn else rsp[i]) & 7
        mod = sgn if mag%3 == 1 else 1-sgn
        if lut == lut1 and  (rsp[i]==1 or rsp[i]==9):
            mod = 1
        if lut == lut1 and  (rsp[i]==11 or rsp[i]==14):
            mod = 0
        lut = lut1 if mod==1 else lut0

def Yoshiencdec(enc, seed=0x85, multiplier=0x260BCD5):
    """Decrypts the IPL3 bootcode copied to rdram."""
    from array import array
    key = multiplier * seed
    key += 1
    m = array("L", enc)
    m.byteswap()
    for i in range(len(m)):
        key &= 0xFFFFFFFF
        m[i] ^= key
        key *= multiplier
    m.byteswap()
    return m.tobytes()

def DDdecrypt(enc):
    """Decrypts the boot sector of a combo 64DD disk.
    Returns decoded data as a bytes object.

    If more than 0x100 bytes are provided unencrypted portion is appended to output."""
    dec = bytearray(0x100)
    count = 0
    # Bytes are copied at 17 byte intervals
    for i in range(0x100):
        dec[count] = enc[i]
        count+=17
        count&=0xFF
    return b''.join((dec, enc[0x100:]))

def DDencrypt(dec):
    """Decrypts the boot sector of a combo 64DD disk.
    Returns decoded data as a bytes object.

    If more than 0x100 bytes are provided unencrypted portion is appended to output."""
    enc = bytearray(0x100)
    count = 0
    # Bytes are copied at 17 byte intervals
    for i in range(0x100):
        enc[i] = dec[count]
        count+=17
        count&=0xFF
    return b''.join((enc, dec[0x100:]))

def RebootDecrypt(enc, func):
    """Decrypts encoded ASM enc using caller rdram address of func as a key,
        returning a bytes object.
    Internally, first set of data is expected to be 0x324 long, second set 0x60."""
    key = func + (func>>8) + (func>>16) + (func>>24)
    key&= 0xFF
    b = bytearray(enc)
    for i in range(0, len(enc), 4):
        b[i+2]= (b[i+2] - key) & 0xFF
        b[i+3]= (b[i+3] + key) & 0xFF
    return bytes(b)

# TODO: roll into a Disk() class with...things.
def getDiskInfo(data):
    """Figures out if you have a dev or retail disk and extracts disk type.
    Returns a tuple: (disk type, device type, boot address, #LBAs loaded)"""
    if data[0x18:0x20] == b'\xff\xff\xff\xff\x00\x00\x00\x00':
        pos = LBA(0).offset(2)  # Same offset regardless disk type.
        d = 'dev'
    else:
        pos = 0
        d = 'retail'
    t = data[pos+5] & 0xF
    n = int.from_bytes(data[6:8], 'big')
    a = int.from_bytes(data[28:32], 'big')
    return (t, d, a, n)

class LBA():
    pzone = ((0,1,2,14,15,3,4,5,6,7,8,9,10,11,12,13),
            (0,1,2,3,13,14,15,4,5,6,7,8,9,10,11,12),
            (0,1,2,3,4,12,13,14,15,5,6,7,8,9,10,11),
            (0,1,2,3,4,5,11,12,13,14,15,6,7,8,9,10),
            (0,1,2,3,4,5,6,10,11,12,13,14,15,7,8,9),
            (0,1,2,3,4,5,6,7,9,10,11,12,13,14,15,8),
            (0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15))
    lbas  = ((0x124, 0x248, 0x35A, 0x47E, 0x5A2, 0x6B4, 0x7C6, 0x8D8, 0x9EA, 0xAB6, 0xB82, 0xC94, 0xDA6, 0xEB8, 0xFCA, 0x10DC),
            (0x124, 0x248, 0x35A, 0x46C, 0x57E, 0x6A2, 0x7C6, 0x8D8, 0x9EA, 0xAFC, 0xBC8, 0xC94, 0xDA6, 0xEB8, 0xFCA, 0x10DC),
            (0x124, 0x248, 0x35A, 0x46C, 0x57E, 0x690, 0x7A2, 0x8C6, 0x9EA, 0xAFC, 0xC0E, 0xCDA, 0xDA6, 0xEB8, 0xFCA, 0x10DC),
            (0x124, 0x248, 0x35A, 0x46C, 0x57E, 0x690, 0x7A2, 0x8B4, 0x9C6, 0xAEA, 0xC0E, 0xD20, 0xDEC, 0xEB8, 0xFCA, 0x10DC),
            (0x124, 0x248, 0x35A, 0x46C, 0x57E, 0x690, 0x7A2, 0x8B4, 0x9C6, 0xAD8, 0xBEA, 0xD0E, 0xE32, 0xEFE, 0xFCA, 0x10DC),
            (0x124, 0x248, 0x35A, 0x46C, 0x57E, 0x690, 0x7A2, 0x86E, 0x980, 0xA92, 0xBA4, 0xCB6, 0xDC8, 0xEEC, 0x1010, 0x10DC),
            (0x124, 0x248, 0x35A, 0x46C, 0x57E, 0x690, 0x7A2, 0x86E, 0x93A, 0xA4C, 0xB5E, 0xC70, 0xD82, 0xE94, 0xFB8, 0x10DC))
    blocks= (0x4D08, 0x47B8, 0x4510, 0x3FC0, 0x3A70, 0x3520, 0x2FD0, 0x2A80, 0x2530)
    track = (0xE8, 0xD8, 0xD0, 0xC0, 0xB0, 0xA0, 0x90, 0x80, 0x70)
    # ROM zones, RAM zones, LBA start of RAM, size of RAM.
    ramrom= ((2,6,0x5A2,0x24A9DC0),
            (3,5,0x7C6,0x1C226C0),
            (4,4,0x9EA,0x1450F00),
            (5,3,0xC0E,0xD35680),
            (6,2,0xE32,0x6CFD40),
            (7,1,0x1010,0x1DA240),
            (8,0,0x10DC,0))

    def __init__(self, disk):
        self.disk = disk
        self.ram = self.ramrom[disk][2]

    def __toIndex(self, lba):
        for i in range(16):
            if lba < self.lbas[self.disk][i]:
                return i
        return -1

    def toVzone(self, lba):
        z = self.toPzone(lba)
        return self.Pzone2Vzone(z)

    def toPzone(self, lba):
        i = self.__toIndex(lba)
        return -1 if i<0 else self.pzone[self.disk][i]

    @staticmethod
    def Pzone2Vzone(zone):
        if zone > 8: zone = 16 - zone
        return zone

    @staticmethod
    def itoPzone(index):
        return LBA.pzone[i] if 0<=i<16 else -1

    @staticmethod
    def itoVzone(index):
        return LBA.Pzone2Vzone(LBA.itoPzone(index))

    def toBytes(self, lba, length, base=24):
        lba += base
        length+=lba
        size = 0
        p = lba
        # Determine initial lba range.
        i = self.__toIndex(lba)
        if i<0:
            raise ValueError("unmappable LBA {:d} (0x{:X}).".format(lba, lba))
        # Calc size of each zone.
        while p < length:
            u = min(self.lbas[self.disk][i], length)
            b = self.toVzone(p)
            size += self.blocks[b] * (u - p)
            p = u
            i+= 1
        return size

    def offset(self, lba):
        """Returns the offset to [lba] for the given disk type."""
        if lba not in range(0x10DD):
            raise ValueError("LBA must be between 0 and 0x10DC")
        return self.toBytes(0, lba, 0)

    def closest(self, offset):
         i = 0
         while self.offset(i) <= offset:
             i+=1
         o = self.offset(max(0, i-1))
         print("LBA 0x{:X}: 0x{:X} + 0x{:X}".format(max(0, i-1), o, offset - o))


# Disk System area:
# 0x0   4
# 0x4   1   version?
# 0x5   1   disk type; mask for actual
# 0x6   2   #blocks to load at LBA "0"
# 0x18  8   boot rdram address
# 0xE0  2   ROM LBA end
# 0xE2  2   RAM LBA start
# 0xE4  2   RAM LBA end

class Cart():
    cic_names = {
        "starf":0x3F,
        "lylat":0x3F,
        "mario":0x3F,
        "diddy":0x78,
        "aleck":0xAC,
        "zelda":0x91,
        "yoshi":0x85,
        "ddipl":0xDD,
        }

    def __init__(self, rom):
        """Sets ROM data, unswaps it if swapped,
            and extracts some handy header data.

        Expects rom to be a byte-like object."""
        self.rom = bytearray(rom)
        self.straighten()
        self.ID = header(rom[0:64])

    def __str__(self):
        """Returns the ROM's internal name, followed by gameID code and version number if applicable."""
        i = self.ID
        return "{}-{}{}{}{}".format(i.title, i.frmt, i.name, i.region, '' if not i.version else str(i.version))

    def cmp(self, **kwargs):
        """Compares named attributes in rom ID with kwargs.
        If all attributes match kwargs, return True.
        Typically used to test if the game ID, region, or version match.
        """
        for i in kwargs.keys():
            if getattr(self.ID, i, None)!=kwargs[i]:
                return False
        return True

    @staticmethod
    def cic2seed(cic):
        # Normalize the names.  (kinda makes the dict pointless...)
        cic = cic.lower()
        if cic in ("6102", "7101", "mario"):
            cic = "mario"
        elif cic in ("6101", "starf"):
            cic = "starf"
        elif cic in ("7102", "lylat"):
            cic = "starf"
        elif cic in ("6103", "7103", "diddy"):
            cic = "diddy"
        elif cic in ("6105", "7105", "zelda"):
            cic = "zelda"
        elif cic in ("6106", "7106", "yoshi"):
            cic = "yoshi"
        elif cic in ("5101", 'aleck'):
            cic = 'aleck'
        elif cic in ("8303", "ddipl"):
            cic = "ddipl"
        else:
            raise TypeError("Unknown CIC type {}.".format(cic))
        return cic, Cart.cic_names.get(cic)

    def crc(self):
        """Returns CRC in rom as a tuple of integers."""
        u = int.from_bytes(self.rom[16:20], byteorder='big')
        l = int.from_bytes(self.rom[20:24], byteorder='big')
        return (u, l)


    def calccrc(self, cic='mario', fix=False, seed=None, base=0x1000, seel=None):
        """Recalculates the CRC based on the CIC chip version given.
        Set fix to True to revise the crc in self.rom.

        cic types:
            'aleck', '5101'
            'starf', '6101'
            'lylat', '7102'
            'mario', '6102', '7101'
            'diddy', '6103', '7103'
            'ddipl', '8303'
            'zelda', '6105', '7105'
            'yoshi', '6106', '7106'
        Setting seed overrides the normal seed value for these types.
        If you need to use one of the particular cic algorithms set the
            string version of the name in cic.  Now that the Aleck64
            CIC has been identified this will probably not be necessary
            unless some repro cic chips start using different values."""
        def rol(v, n):
            return (v % 0x100000000)>>n

        # The algo slightly changes with certain cics.
        cic, s = self.cic2seed(cic)
        # Pull the seed byte and generate the seed from it.
        # Note even if seed is set you'll have to pass a cic "name" to get funky types out.
        if seed is not None:
            s = seed
        if cic in ('diddy', 'yoshi', 'aleck'):
            seed = 0x6C078965 * s
        elif cic == 'ddipl':
            seed = 0x2E90EDD * s
        else:
            seed = 0x5D588B65 * s
        seed+= 1
        seed&=0xFFFFFFFF
        r1, r2, r3, r4, r5, r6 = seed, seed, seed, seed, seed, seed

        # I wish there was a less horrifying way to do this...
        from array import array
        if seel is None:
            if cic == 'aleck':
                seel = 0x3FE000 if self.ID.PC == 0x80100400 else 0x100000
            elif cic == 'ddipl':
                seel = 0xA0000
            else:
                seel = 0x100000
        seel += base
        l = min(seel, len(self.rom))
        m = array("L", self.rom[base:l] + bytes(seel - l))
        m.byteswap()
        # Zelda updates the second word a different way...
        if cic == 'zelda':
            from itertools import cycle
            n = array("L", self.rom[0x750:0x850])
            n.byteswap()
            n = cycle(n)
        # Read each word as an integer.
        for i in m:
            v = (r1+i) & 0xFFFFFFFF
            if v < r1:
                if cic == 'ddipl':
                    r2^=r3
                else:
                    r2+=1
            v = i & 0x1F
            a = (i<<v) | (rol(i, 0x20-v))
            r1+=i
            r3^=i
            r4+=a
            # You have to limit the result here to 32bits.
            r1&= 0xFFFFFFFF
            r4&= 0xFFFFFFFF
            if r5 < i:
                r5^= (r1^i)
            else:
                if cic == 'ddipl':
                    r5+=a
                else:
                    r5^=a
            if cic == 'zelda':
                r6+= (i ^ next(n))
            else:
                r6+= (i ^ r4)
            # Ditto here.
            r5&= 0xFFFFFFFF
            r6&= 0xFFFFFFFF
        # Assemble upper and lower CRCs
        if cic == 'ddipl':
            if fix:
                if isinstance(self.rom, bytes):
                    self.rom = bytearray(self.rom)
                self.setASMvalue(r1, 0x608, 0x60C)
                self.setASMvalue(r2, 0x618, 0x61C)
                self.setASMvalue(r3, 0x628, 0x62C)
                self.setASMvalue(r4, 0x638, 0x63C)
                self.setASMvalue(r5, 0x648, 0x64C)
                self.setASMvalue(r6, 0x658, 0x65C)
            return (r1, r2, r3, r4, r5, r6)
        if cic == 'yoshi':
            r1*=r2
            r4*=r5
        else:
##            r2&= 0xFFFFFFFF
            r1^=r2
            r4^=r5
        if cic in ('diddy', 'yoshi', 'aleck'):
            r1+=r3
            r4+=r6
        else:
            r1^=r3
            r4^=r6
        # Make sure they fit within 4 bytes each.
        r1&= 0xFFFFFFFF
        r4&= 0xFFFFFFFF
        if fix:
            if isinstance(self.rom, bytes):
                self.rom = bytearray(self.rom)
            self.rom[16:20] = r1.to_bytes(4, 'big')
            self.rom[20:24] = r4.to_bytes(4, 'big')
        return (r1,r4)

    def bootstrapcrc(self, cic='mario', seed=None):
        import struct
        from array import array

        if seed is None:
            cic, seed = self.cic2seed(cic)

        def trimult(v1, v2, v3):
            if not v2:
                v2 = v3
            l = v1 * v2
            u = (l>>32) & 0xFFFFFFFF
            l &= 0xFFFFFFFF
            u -= l
            if not u:
                u = v1
            return (u & 0xFFFFFFFF)

        def rol(v, n):
            return (v % 0x100000000)>>n

        data = struct.unpack(">1008L", self.rom[0x40:0x1000])

        seed *= 0x6C078965
        seed += 1
        seed ^= data[0]
        seed &= 0xFFFFFFFF
        regs = [seed, seed, seed, seed, seed, seed, seed, seed,
                seed, seed, seed, seed, seed, seed, seed, seed,]
        count = 0   #S1
        cur = data[0]  #S4
        # First half of the algorithm: read in bootcode.
        while True:
            prev = cur  #S4
            cur = data[count]   #S0
            count += 1
            v = (0x3EF - count) & 0xFFFFFFFF
            regs[0] +=  trimult(v, cur, count)
            regs[0] &= 0xFFFFFFFF
            regs[1] = trimult(regs[1], cur, count)
            regs[2] ^= cur
            regs[3] += trimult(cur + 5, 0x6C078965, count)
            regs[3] &= 0xFFFFFFFF
            # BFC00250
            if prev < cur:
                regs[9] = trimult(regs[9], cur, count)
            else:
                regs[9] += cur
            regs[9] &= 0xFFFFFFFF
            # BFC00288
            b = prev & 0x1F
            u = cur << (32 - b)
            l = rol(cur, b)
            roll1 = (u | l) & 0xFFFFFFFF   #S5
            regs[4] += roll1
            regs[4] &= 0xFFFFFFFF
            b = prev & 0x1F
            u = cur << b
            l = rol(cur, 32 - b)
            v = (u | l) & 0xFFFFFFFF
            regs[7] = trimult(regs[7], v, count)
            # BFC002C0
            if cur < regs[6]:
                regs[6] += regs[3]
                regs[6] ^= cur + count
            else:
                regs[6] ^= regs[4] + cur
            regs[6] &= 0xFFFFFFFF
            # BFC002FC
            b = prev >> 0x1B
            u = cur << b
            l = rol(cur, 32 - b)
            roll2 = (u | l) & 0xFFFFFFFF   #S2
            regs[5] += roll2
            regs[5] &= 0xFFFFFFFF
            u = cur << (32 - b)
            l = rol(cur, b)
            v = (u | l) & 0xFFFFFFFF
            regs[8] = trimult(regs[8], v, count)
            if count == 0x3F0:
                break
            future = data[count]    #S3
            # BFC00340
            regs[15] = trimult(regs[15], roll2, count)
            b = rol(cur, 0x1B)
            u = future << b
            l = rol(future, 32 - b)
            v = (u | l) & 0xFFFFFFFF
            regs[15] = trimult(regs[15], v, count)
            # BFC00374
            regs[14] = trimult(regs[14], roll1, count)
            b = cur & 0x1F  #S2
            u = future << (32 - b)
            l = rol(future, b)
            v = (u | l) & 0xFFFFFFFF
            regs[14] = trimult(regs[14], v, count)
            # BFC003A4
            u = cur << (32 - b)
            l = rol(cur, b)
            v = (u | l) & 0xFFFFFFFF
            regs[13] += v
            b = future & 0x1F
            u = future << (32 - b)
            l = rol(future, b)
            v = (u | l) & 0xFFFFFFFF
            regs[13] += v
            regs[13] &= 0xFFFFFFFF
            # BFC003DC
            regs[10] += cur
            regs[10] = trimult(regs[10] & 0xFFFFFFFF, future, count)
            regs[11] = trimult(regs[11] ^ cur, future, count)
            regs[12] += regs[8] ^ cur
            regs[12] &= 0xFFFFFFFF
        # BFC00420
        buf = [regs[0], regs[0], regs[0], regs[0]]
        for i in range(16):
            cur = regs[i]
            b = cur & 0x1F
            u = cur << (32 - b)
            l = rol(cur, b)
            v = (u | l) & 0xFFFFFFFF
            buf[0] += v
            buf[0] &= 0xFFFFFFFF
            if cur < buf[0]:
                buf[1] += cur
                buf[1] &= 0xFFFFFFFF
            else:
                buf[1] = trimult(buf[1], cur, i)
            # BFC00494
            if (cur & 3) in (0, 3):
                buf[2] += cur
                buf[2] &= 0xFFFFFFFF
            else:
                buf[2] = trimult(buf[2], cur, i)
            # BFC004CC
            if cur & 1:
                buf[3] ^= cur
            else:
                buf[3] = trimult(buf[3], cur, i)
        # BFC00504
        v = trimult(buf[0], buf[1], 16)
        b = buf[2] ^ buf[3]
        return (v & 0xFFFFFFFF, b & 0xFFFFFFFF)

    def byteorder(self):
        return byteorder(self.rom[0:4])

    def straighten(self):
        """Sets self.rom big endian, byteswapping if necessary."""
        if self.byteorder() == 'little':
            self.rom = byteswap(self.rom)

    def getASMvalue(self, upper, lower=None):
        if lower is None:
            u = int.from_bytes(self.rom[upper+2:upper+4], 'big') & 0x3FFFFFF
            u<<=2
            return u | 0x80000000
        u = int.from_bytes(self.rom[upper+2:upper+4], 'big')<<16
        e = False if self.rom[lower]&0x10 else True
        l = int.from_bytes(self.rom[lower+2:lower+4], 'big', signed=e)
        return u+l

    def setASMvalue(self, value, upper, lower=None):
        if lower is None:
            value>>=2
            value&=0x3FFFFFF
            u = self.rom[upper] & 0xFC
            value|= u<<24
            self.rom[upper:upper+4] = value.to_bytes(4, 'big')
        else:
            u = value>>16
            l = value&0xFFFF
            if not self.rom[lower]&0x10:
                u += bool(l&0x8000)
            self.rom[upper+2:upper+4] = u.to_bytes(2, 'big')
            self.rom[lower+2:lower+4] = l.to_bytes(2, 'big')

class MFS:
    def __init__(self, disk, mngr1, mngr2, data):
        from collections import namedtuple

        mngr = namedtuple("Mngr", "header, tbl, entries")
        self.disk = disk
        self.data = data
        self.lba = LBA(disk)
        self.mngr1 = mngr(None, self.header(mngr1), list())
        for i in range(0x16B0, len(mngr1), 48):
            if not mngr1[i]:
                continue
            self.mngr1.entries.append(self.entry(mngr1[i:i+48]))

        self.mngr2 = mngr(None, self.header(mngr2), list())
        for i in range(0x16B0, len(mngr2), 48):
            if not mngr2[i]:
                continue
            self.mngr2.entries.append(self.entry(mngr2[i:i+48]))

    @classmethod
    def fromdisk(cls, data):
        disk = getDiskInfo(data)[0]
        lba = LBA(disk)
        start = lba.offset(lba.ram)
        end = lba.offset(0x10DC)
        return cls.frombytes(disk, data[start:end])

    @classmethod
    def frombytes(cls, disk, data):
        lba = LBA(disk)
        mngr1end = lba.toBytes(lba.ram, 3, 0)
        mngr1 = data[0:mngr1end]
        mngr2end = lba.toBytes(lba.ram, 6, 0)
        mngr2 = data[mngr1end:mngr2end]
        return cls(disk, mngr1, mngr2, data[mngr2end:])

    @staticmethod
    def header(data):
        from array import array
        tbl = array("h", data[0x3C:0x16B0])
        tbl.byteswap()
        return tbl

    @staticmethod
    def entry(data):
        from collections import namedtuple

        Entry = namedtuple("Entry", 'attr, parent, company, gameID, fat, filesz, name, ext, copycnt, renewcnt, date')

        e = int.from_bytes(data[0:2], 'big')
        attr = {"dir": bool(e & 0x8000),
                "file":bool(e & 0x4000),
                "rdprotect":bool(e & 0x2000),
                "wrprotect":bool(e & 0x1000),
                "hidden":bool(e & 0x800),
                "encoded":bool(e & 0x400),
                "copycnt":bool(e & 0x200),
                }
        d = int.from_bytes(data[2:4], 'big', signed=True)
        f = int.from_bytes(data[10:12], 'big')
        if attr['file']:
            sz= int.from_bytes(data[12:16], 'big')
            ext = data[36:41].decode('shift_jisx0213').strip('\x00')
            c = data[41]
        else:
            sz = 0
            ext = ''
            c = 0
        name = data[16:36].decode('shift_jisx0213').strip('\x00')
        e = int.from_bytes(data[44:48], 'big')
        date = {'year':(e>>25) + 1996,
                'month':(e>>21)&0xF,
                'day':(e>>16)&0x1F,
                'hr':(e>>11)&0x1F,
                'min':(e>>5)&0x3F,
                'sec':e&0x1F,
                }
        return Entry(attr, d, data[4:6].decode(), data[6:10].decode(), f, sz, name, ext, c, data[42], date)

    def extract(self, index):
        e = self.mngr1.entries[index]
        l = self.mngr1.tbl[e.fat]
        o = self.lba.toBytes(self.lba.ram + 6, l - 6, 0)
        return self.data[o:o + e.filesz]

    @staticmethod
    def _checksum(mngr):
        """Checksum computed from file management area.
        Blank out checksum field and do a 32bit XOR on all values."""
        from array import array
        a = array("L", mngr)
        a.byteswap()
        a[11] = 0
        c = 0
        for i in a:
            c ^= i
        return c

def main():
    import argparse
    parser = argparse.ArgumentParser(description="Calculates checksum of <file> using <cic>.")
    parser.add_argument('cic', type=str, nargs='?', default=None)
    parser.add_argument('file', type=argparse.FileType('rb'))
    cl = parser.parse_args()
    if cl.file is None: return

    rom = Cart(cl.file.read())
    cl.file.close()
    if cl.cic is None:
        cic = {
        0:"starf",
        0xD5:"lylat",
        0xDE:"mario",
        0xDB:"diddy",
        0xE4:"aleck",
        0x14:"zelda",
        0xEC:"yoshi",
        }.get(rom.rom[0x173], 'mario')
        if cic == "starf" and rom.rom[0x16F]==0xDB:
            cic = "ddipl"
    else:
        cic = cl.cic
    cic = Cart.cic2seed(cic)
    if cic[0] == 'ddipl':
        print("internal\t{:08X} {:08X} {:08X} {:08X} {:08X} {:08X}".format(rom.getASMvalue(0x608, 0x60C), rom.getASMvalue(0x618, 0x61C), rom.getASMvalue(0x628, 0x62C), rom.getASMvalue(0x638, 0x63C), rom.getASMvalue(0x648, 0x64C), rom.getASMvalue(0x658, 0x65C)))
        print("calculated\t{:08X} {:08X} {:08X} {:08X} {:08X} {:08X}".format(*rom.calccrc(cic[0])))
    else:
        print("internal\t{:08X} {:08X}".format(*rom.crc()))
        print("calculated\t{:08X} {:08X}".format(*rom.calccrc(cic[0])))

if __name__ == '__main__':
    main()
